/* FLENG - runtime system, C part */

#include "fleng.h"
#include "fleng-util.h"
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <assert.h>
#include <stdarg.h>
#include <unistd.h>
#include <errno.h>
#include <math.h>
#include <stdint.h>
#include <sys/time.h>
#ifdef __linux__
#include <alloca.h>
#endif

#if defined(__LP64__) || defined(_LP64)
#define MAX_THREADS     64
#else
#define MAX_THREADS     32
#endif
#define MAX_SSTACK      256
#define MAX_ENV         64
#define TIMESLICE       10
#define STAT_TICKS      50
#define MAX_DELAY       (100 * 256)
#define PORT_SIZE       0x1000

/* XXX */
#define FUDGE           10

/* [SEND_ADDREF, Rptr(Owner, x)] */
#define SEND_ADDREF     1
/* [SEND_DROPREF, Rptr(Owner, x)] */
#define SEND_DROPREF    2
/* [SEND_SENDPORT, Rptr(Owner, Port)|x] */
#define SEND_SENDPORT   3
/* [SEND_READ, Rptr(Owner, Var)|NewVar] */
#define SEND_READ       4
/* [SEND_VALUE, Rptr(Owner, Var)|Val] */
#define SEND_VALUE      5
/* [SEND_CALL, ID, Term|Mod] */
#define SEND_CALL       6

#define HEAP_SIZE       1000000
#define MAX_GOALS       10000

#define THREAD_BIT(tcb)   ((long)(1 << (tcb->ordinal - 1)))

FL_VAL fl_nil, fl_true, fl_false, fl_colon, sys_name;
FL_TCB fl_tcbs[ MAX_THREADS ];
static int ntcbs = 0, logging = 0;
static char *strings = NULL;
char *fl_argv[ 256 ];
static FILE *logfile;
static int stats = 0, heapstats = 0, stat_ticks = STAT_TICKS;
static int threads;
static volatile long idle_threads;
static pthread_mutex_t log_mutex = PTHREAD_MUTEX_INITIALIZER;
static pthread_mutex_t intern_mutex = PTHREAD_MUTEX_INITIALIZER;
static FL_MODULE *last_module = NULL;
static int timeslice = TIMESLICE;

static void terminate(int code)
{
    fflush(logfile);
    exit(code);
}

int fl_nthreads(void) { return threads; }

void fl_write(FL_TCB *tcb, FILE *fp, FL_VAL x)
{
    x = deref(x);
    if(((long)x & FL_INT_BIT) != 0)
        fprintf(fp, "%ld", (long)x >> 1);
    else if(((long)x & FL_ATOM_BIT) != 0)
        fputs((char *)x + 1, fp);
    else {
        FL_CELL *c = (FL_CELL *)x;
        unsigned long tag = c->tag;
        if(c->car == (FL_VAL)FL_BROKEN_HEART) {
            fprintf(fp, "<RECLAIMED %p>", x);
            return;
        }
        if((tag & FL_TUPLE_BIT) != 0) {
            FL_VAL x1 = c->car;
            int a = (tag >> 24) & 0x7f;
            if(((long)x1 & FL_BITS_MASK) == FL_ATOM_BIT) {
                fl_write(tcb, fp, x1);
                fputc('(', fp);
                x = c->cdr;
                fl_write(tcb, fp, ((FL_CELL *)x)->car);
                for(x1 = ((FL_CELL *)x)->cdr;
                    x1 != fl_nil; 
                    x1 = ((FL_CELL *)x1)->cdr) {
                    fputs(", ", fp);
                    fl_write(tcb, fp, ((FL_CELL *)x1)->car);
                }
                fputc(')', fp);
            } else {
                fputc('{', fp);
                fl_write(tcb, fp, x1);
                x1 = c->cdr; 
                while(x1 != fl_nil) {
                    fputs(", ", fp);
                    fl_write(tcb, fp, ((FL_CELL *)x1)->car);
                    x1 = ((FL_CELL *)x1)->cdr;
                }
                fputc('}', fp);
            }
        } else {
            switch(tag & FL_TAG_MASK) {
            case FL_FLOAT_TAG:
                fprintf(fp, "%.*g", 15, *((double *)(&c->car)));
                break;
            case FL_LIST_TAG:
                fputc('[', fp);
                fl_write(tcb, fp, c->car);
                x = deref(c->cdr); 
                while(((long)x & FL_BITS_MASK) == 0 && TAG(x) == FL_LIST_TAG) {
                    fputs(", ", fp);
                    fl_write(tcb, fp, ((FL_CELL *)x)->car);
                    x = deref(((FL_CELL *)x)->cdr);
                }
                if(x != fl_nil) {
                    fputc('|', fp);
                    fl_write(tcb, fp, x);
                }
                fputc(']', fp);
                break;
            case FL_VAR_TAG:
                fprintf(fp, "_%lu", (unsigned long)x);
                break;
            case FL_PORT_TAG:
                if(((long)c->car & FL_INT_BIT) != 0)
                    fprintf(fp, "<port %p@%ld>", RPTR(c->car), RPTR_ID(c->car));
                else fprintf(fp, "<port %p>", c);
                break;
            case FL_REF_TAG:
                x = CAR(x);
                fprintf(fp, "_%ld_%ld", (long)RPTR(x), RPTR_ID(x));
                break;
            case FL_MODULE_TAG:
                fprintf(fp, "<module %s>", (char *)((FL_MODULE *)((long)CAR(x) & ~1))->name + 1);
                break;
            default:
                fprintf(fp, "<unknown cell with tag 0x%04lx>", tag);
                break;
            }
        }
    }
}

void fl_logfmt(FL_TCB *tcb, char *fstr, ...)
{
    va_list va;
    va_start(va, fstr);
    char *info = NULL;
    if(tcb->goal != NULL) info = tcb->goal->info;
    if(info == NULL) info = "<none>";
    pthread_mutex_lock(&log_mutex);
    fprintf(logfile, "<%d> %ld: %s ", tcb->ordinal, tcb->ticks, info);
    vfprintf(logfile, fstr, va);
    pthread_mutex_unlock(&log_mutex);
    va_end(va);    
}

void fl_log_1(FL_TCB *tcb, FL_VAL msg)
{
    if(tcb->logging) {
        pthread_mutex_lock(&log_mutex);
        FL_GOAL *g = tcb->goal;
        fprintf(logfile, "<%d> %ld: ", tcb->ordinal, tcb->ticks);
        fl_write(tcb, logfile, msg);
        fputc('\n', logfile);
        pthread_mutex_unlock(&log_mutex);
    }
}

void fl_log_entry(FL_TCB *tcb, long arity)
{
    if(tcb->logging) {
        pthread_mutex_lock(&log_mutex);
        FL_GOAL *g = tcb->goal;
        fprintf(logfile, "<%d> %ld: %s ", tcb->ordinal, tcb->ticks, g->info);
        int i;
        for(i = 0; i < arity; ++i) {
            fl_write(tcb, logfile, g->args[ i ]);
            fputc(' ', logfile);
            if(i == 2 && arity > 4) {
                for(FL_VAL a = g->args[ 3 ]; 
                    a != fl_nil; 
                    a = ((FL_CELL *)a)->cdr) {
                    fl_write(tcb, logfile, ((FL_CELL *)a)->car);
                    fputc(' ', logfile);
                }
                break;
            }
        }
        fputc('\n', logfile);
        pthread_mutex_unlock(&log_mutex);
    }
}

static void statistics(FL_TCB *tcb)
{
    pthread_mutex_lock(&log_mutex);
    fprintf(logfile, "## %d %ld G %ld Q %ld S %ld C %ld\n", tcb->ordinal, 
        tcb->ticks, tcb->goals, tcb->active, tcb->suspended, tcb->used);
    pthread_mutex_unlock(&log_mutex);
}

static void heap_statistics(FL_TCB *tcb)
{
   long vars = 0, tuples = 0, lists = 0;
   for(FL_CELL *p = tcb->heap; p < tcb->heapend; ++p) {
        switch(p->tag & FL_TAG_MASK) {
        case FL_LIST_TAG: ++lists; break;
        case FL_VAR_TAG: ++vars; break;
        default:
            if((p->tag & FL_TUPLE_BIT) != 0) ++tuples;
        }
    }
    tuples *= sizeof(FL_CELL);
    lists *= sizeof(FL_CELL);
    vars *= sizeof(FL_CELL);
    pthread_mutex_lock(&log_mutex);
    fprintf(logfile, "#$ %d %ld T %ld L %ld V %ld\n", tcb->ordinal, tcb->ticks,
        tuples / 1024, lists / 1024, vars / 1024);
    pthread_mutex_unlock(&log_mutex);
}

void fl_rt_error(FL_TCB *tcb, FL_VAL x, int error)
{
    fprintf(stderr, "\n*** ERROR: <%d> %ld: %s - ", tcb->ordinal, tcb->ticks, 
        tcb->goal->info);
    switch(error) {
    case FL_NOT_AN_ATOM:
        fputs("expected atom but received: ", stderr);
        fl_write(tcb, stderr, x);
        break;
    case FL_NOT_A_CELL:
        fputs("expected cell but received: ", stderr);
        fl_write(tcb, stderr, x);
        break;
    case FL_NOT_AN_INT:
        fputs("expected integer but received: ", stderr);
        fl_write(tcb, stderr, x);
        break;
    case FL_NOT_A_LIST:
        fputs("expected list but received: ", stderr);
        fl_write(tcb, stderr, x);
        break;
    case FL_NOT_A_TUPLE:
        fputs("expected tuple but received: ", stderr);
        fl_write(tcb, stderr, x);
        break;
    case FL_NOT_A_VAR:
        fputs("expected variable but received: ", stderr);
        fl_write(tcb, stderr, x);
        break;
    case FL_NOT_A_NUMBER:
        fputs("expected number but received: ", stderr);
        fl_write(tcb, stderr, x);
        break;
    case FL_NOT_A_MODULE:
        fputs("expected module but received: ", stderr);
        fl_write(tcb, stderr, x);
        break;
    case FL_QUEUE_FULL:
        fputs("goal queue overflow", stderr);
        break;
    case FL_NO_MODULE:
        fputs("module not found: ", stderr);
        fl_write(tcb, stderr, x);
        break;
    case FL_NO_PDEF:
        fputs("process definition not found: ", stderr);
        fl_write(tcb, stderr, x);
        break;
    case FL_NO_CELLS:
        fputs("out of memory", stderr);
        break;
    case FL_DIV_BY_ZERO:
        fputs("division by zero", stderr);
        break;
    case FL_CANT_CONVERT:
        fputs("can not convert value: ", stderr);
        fl_write(tcb, stderr, x);
        break;
    case FL_NOT_A_STRING:
        fputs("expected string or character list: ", stderr);
        fl_write(tcb, stderr, x);
        break;
    case FL_IO_ERROR:
        fprintf(stderr, "I/O error - %s: ", strerror(errno));
        fl_write(tcb, stderr, x);
        break;
    case FL_NOT_A_PORT:
        fputs("expected port: ", stderr);
        fl_write(tcb, stderr, x);
        break;
    case FL_PORT_CLOSED:
        fputs("sent value to closed port: ", stderr);
        fl_write(tcb, stderr, x);
        break;
    case FL_BAD_INDEX:
        fputs("index out of range: ", stderr);
        fl_write(tcb, stderr, x);
        break;
    case FL_CANT_FORWARD:
        fputs("unable to forward object: ", stderr);
        fl_write(tcb, stderr, x);
        break;
    case FL_CANT_UNFORWARD:
        fputs("unable to unforward object", stderr);
        break;
    case FL_BAD_PEER:
        fputs("invalid peer: ", stderr);
        fl_write(tcb, stderr, x);
        break;
    case FL_OUT_OF_MEMORY:
        fputs("out of memory", stderr);
        break;
    case FL_BAD_SIGNAL:
        fputs("invalid signal: ", stderr);
        fl_write(tcb, stderr, x);
        break;
    case FL_TOO_MANY_GOALS:
        fputs("goal limit exceeded\n", stderr);
        break;
    default:
        fprintf(stderr, "unknown error code %d", error);
    }
    fputc('\n', stderr);
    terminate(1);
}

void fl_rt_error2(FL_TCB *tcb, FL_VAL x, FL_VAL y, int error)
{
    fprintf(stderr, "<%d> %ld: %s - ", tcb->ordinal, tcb->ticks, tcb->goal->info);
    switch(error) {
    case FL_CANT_UNIFY:
        fputs("unification failed:", stderr);
        break;
    default:
        fprintf(stderr, "unknown error code %d", error);
    }
    fprintf(stderr, "\n<%d> %ld: %s   ", tcb->ordinal, tcb->ticks, tcb->goal->info);
    fl_write(tcb, stderr, x);
    fputc('\n', stderr);
    fprintf(stderr, "\n<%d> %ld: %s:   ", tcb->ordinal, tcb->ticks, tcb->goal->info);
    fl_write(tcb, stderr, y);
    fputc('\n', stderr);
    terminate(1);
}

FL_VAL fl_alloc_cell(FL_TCB *tcb, FL_VAL car, FL_VAL cdr)
{
    /* Cell layout: 
        List = [Car Cdr]
        Var = [Val Suspensions(List)]   Val == Var, if unbound
        Float = [Num* Unused^]
        Port = [Stream|Source% Unused]
        Ref = [Id/Var% Unused]
        Module = [Module+ Unused]
        *: in native format
        +: marked with lowest bit set
        %: RPtr 
        ^: only on 64 bit platform */
    if(tcb->freelist == NULL) fl_rt_error(tcb, 0, FL_NO_CELLS);
    FL_CELL *c = tcb->freelist;
    tcb->freelist = tcb->freelist->cdr;
    c->car = addref(car);
    c->cdr = addref(cdr);
    ++tcb->used;
    return (FL_VAL)c;
}

FL_VAL fl_alloc_float(FL_TCB *tcb, double n)
{
#if defined(__LP64__) || defined(_LP64)
    FL_VAL c = fl_alloc_cell(tcb, fl_nil, fl_nil);
#else    
    FL_VAL *prev = (FL_VAL *)&tcb->freelist;
    FL_VAL c = *prev;
    while(((long)c & 4) == 0) {
        if(c == NULL) fl_rt_error(tcb, 0, FL_NO_CELLS);
        prev = &((FL_CELL *)c)->cdr;
        c = *prev;
    }
    *prev = CDR(c);
    ++tcb->used;
#endif
    ((FL_CELL *)c)->tag = FL_FLOAT_TAG;
    *((double *)&((FL_CELL *)c)->car) = n;
    return c;
}

static void enqueue(FL_TCB *tcb, FL_GOAL *g)
{
    *(tcb->qend++) = g;
    if(tcb->qend == tcb->qstart)
        fl_rt_error(tcb, fl_nil, FL_QUEUE_FULL);
    if(tcb->qend >= &tcb->queue[ tcb->max_goals ])
        tcb->qend = tcb->queue;
    ++tcb->active;
}

static FL_GOAL *dequeue(FL_TCB *tcb)
{
    assert(tcb->qstart != tcb->qend);
    FL_GOAL *g = *(tcb->qstart++);
    if(tcb->qstart >= &tcb->queue[ tcb->max_goals ])
        tcb->qstart = tcb->queue;
    --tcb->active;
    return g;
}

void fl_suspend(FL_TCB *tcb)
{
    /* Suspension-list:
       [[GOAL1+, [Var1+, Addr1+], ...], ...]
       +: marked */
    FL_VAL slst = fl_nil;
    for(FL_VAL *p = tcb->sstack; p < tcb->sstackp; p += 2) {
        FL_VAL var = *p;
        FL_VAL addr = (FL_VAL)((long)p[ 1 ] | 1);
        if(TAG(var) == FL_REF_TAG) {
            FL_VAL remote = CAR(var);
            ((FL_CELL *)var)->tag = FL_VAR_TAG | COUNT(var);
            ((FL_CELL *)var)->car = var;
            ((FL_CELL *)var)->cdr = fl_nil;            
            FL_VAL ev = mklist(tcb, MKINT(SEND_READ),
                               mklist(tcb, remote, var));
            fl_add_event(tcb, ev);
        }
        FL_VAL x = mklist(tcb, (FL_VAL)((long)var | 1), addr);
        slst = mklist(tcb, x, slst);
    }
    slst = mklist(tcb, (FL_VAL)((long)tcb->goal | 1), slst);
    for(FL_VAL *p = tcb->sstack; p < tcb->sstackp; p += 2) {
        FL_VAL var = *p;
        FL_VAL sl = ((FL_CELL *)var)->cdr;
        ((FL_CELL *)var)->cdr = addref(mklist(tcb, slst, sl));
#ifdef LOGGING
        if(tcb->logging) fl_logfmt(tcb, "suspend on _%ld\n", (long)var);
#endif
        unref(tcb, sl);
    }
    tcb->sstackp = tcb->sstack;
    ++tcb->suspended;
}

static void resume(FL_TCB *tcb, FL_VAL var)
{
    for(FL_VAL x = ((FL_CELL *)var)->cdr; x != fl_nil; x = ((FL_CELL *)x)->cdr) {
        FL_VAL slst = ((FL_CELL *)x)->car;
        if(((FL_CELL *)slst)->car != fl_nil) {
            --tcb->suspended;
            FL_GOAL *g = (FL_GOAL *)((long)((FL_CELL *)slst)->car & ~1);
#ifdef LOGGING
            if(tcb->logging)
                fl_logfmt(tcb, "resuming %s on _%ld\n", g->info, (long)var);
#endif
            enqueue(tcb, g);
            ((FL_CELL *)slst)->car = fl_nil;
            slst = ((FL_CELL *)slst)->cdr;
            while(slst != fl_nil) {
                FL_VAL p1 = ((FL_CELL *)slst)->car;
                FL_VAL v = (FL_VAL)((long)((FL_CELL *)p1)->car & ~1);
                if(v == var) {
                    g->addr = (void *)((long)((FL_CELL *)p1)->cdr & ~1);
                    break;
                }
                slst = ((FL_CELL *)slst)->cdr;
            }
        }
    }
    unref(tcb, ((FL_CELL *)var)->cdr);
    ((FL_CELL *)var)->cdr = fl_nil;
}

FL_VAL fl_chase_tail(FL_VAL st)
{
    st = deref(st);
    while(TAG(st) == FL_LIST_TAG) {
        st = deref(CDR(st));
        assert(((long)st & FL_BITS_MASK) == 0);
    }
    return st;
}

void fl_add_event(FL_TCB *tcb, FL_VAL event)
{
#ifdef LOGGING
    if(tcb->logging) {
        pthread_mutex_lock(&log_mutex);
        fprintf(logfile, "<%d> %ld: %s event: ", tcb->ordinal, tcb->ticks, 
            tcb->goal->info);
        fl_write(tcb, logfile, event);
        fputc('\n', logfile);
        fprintf(logfile, "<%d> %ld: %s   queue: ", tcb->ordinal, tcb->ticks, 
            tcb->goal->info);
        fl_write(tcb, logfile, tcb->events);
        fputc('\n', logfile);
        if(tcb->defer) {
            fprintf(logfile, "<%d> %ld: %s   deferred: ", tcb->ordinal, tcb->ticks,
                tcb->goal->info);
            fl_write(tcb, logfile, tcb->deferred);
            fputc('\n', logfile);
        }
        pthread_mutex_unlock(&log_mutex);
    }
#endif
    FL_VAL var = fl_alloc_cell(tcb, fl_nil, fl_nil);
    ((FL_CELL *)var)->tag = FL_VAR_TAG;
    ((FL_CELL *)var)->car = var;
    FL_VAL n = fl_alloc_cell(tcb, event, var);
    ((FL_CELL *)n)->tag = FL_LIST_TAG;
    if(tcb->defer) {
        fl_assign(tcb, n, tcb->deferred_tail);
        tcb->deferred_tail = var;
        return;
    }
    FL_VAL s = addref(fl_chase_tail(tcb->events));
    assert(BITS(s) == 0 && TAG(s) == FL_VAR_TAG);
    int diddefer = 0;
    if(!tcb->defer) {
        tcb->defer = diddefer = 1;
        tcb->deferred = tcb->deferred_tail = var;
    }
    unref(tcb, tcb->events);
    if(diddefer) tcb->defer = 0;
    fl_assign(tcb, n, s);
    tcb->events = s;
}

void fl_assign(FL_TCB *tcb, FL_VAL x, FL_VAL var)
{
    int isref = TAG(var) == FL_REF_TAG;
#ifdef LOGGING
    if(tcb->logging) {
        pthread_mutex_lock(&log_mutex);
        fprintf(logfile, "<%d> %ld: %s assign ", tcb->ordinal, tcb->ticks, 
            tcb->goal->info);
        if(!isref) fprintf(logfile, "_%ld: ", (long)var);
        else fprintf(logfile, "_%ld_%ld: ", (long)RPTR(CAR(var)), RPTR_ID(CAR(var)));
        fl_write(tcb, logfile, x);
        fputc('\n', logfile);
        pthread_mutex_unlock(&log_mutex);
    }
#endif
    if(isref) {
        FL_VAL ev = mklist(tcb, (FL_VAL)((SEND_VALUE << 1) | 1), 
                        mklist(tcb, CAR(var), addref(x)));
        fl_add_event(tcb, ev);
        ((FL_CELL *)var)->tag = FL_VAR_TAG | (((FL_CELL *)var)->tag & FL_COUNT_MASK);
        CAR(var) = x;
        return;
    }
    if(CAR(var) == x) return;
    if(CAR(var) != var) {
        fprintf(stderr, "<%d> %ld: %s - assignment to non-variable _%ld: ", tcb->ordinal,
            tcb->ticks, tcb->goal->info, (long)var);
        fl_write(tcb, stderr, x);
        fputs(" (is: ", stderr);
        fl_write(tcb, stderr, CAR(var));
        fputs(")\n", stderr);
        terminate(1);
    }
    ((FL_CELL *)var)->car = addref(x);
    if(((FL_CELL *)var)->cdr != fl_nil) resume(tcb, var);
}

static void close_stream(FL_TCB *tcb, FL_VAL st)
{
#ifdef LOGGING
    if(tcb->logging) {
        pthread_mutex_lock(&log_mutex);
        fprintf(logfile, "<%d> %ld: %s closing stream: ", tcb->ordinal, tcb->ticks,
            tcb->goal->info);
        fl_write(tcb, logfile, st);
        fputc('\n', logfile);
        pthread_mutex_unlock(&log_mutex);
    }
#endif
    st = fl_chase_tail(st);
    if(BITS(st) == 0 && TAG(st) == FL_VAR_TAG) 
        fl_assign(tcb, fl_nil, st);
}

void fl_release(FL_TCB *tcb, FL_VAL x)
{
#ifdef XXLOGGING
    if(tcb->logging) {
        pthread_mutex_lock(&log_mutex);
        fprintf(logfile, "<%d> %ld: %s release %p (0x%08lx) [%ld]: ", tcb->ordinal, 
            tcb->ticks, tcb->goal->info, x, TAG(x), tcb->used);
        fl_write(tcb, logfile, x);
        fputc('\n', logfile);
        pthread_mutex_unlock(&log_mutex);
    }
#endif
    FL_CELL *c = (FL_CELL *)x;
    FL_VAL tl;
    void *addr;
    switch(c->tag & FL_TAG_MASK) {
    case FL_FLOAT_TAG: break;
    case FL_VAR_TAG:
        if(c->car != x) unref(tcb, c->car);
        unref(tcb, c->cdr);
        break;
    case FL_PORT_TAG:
        if(c->car == fl_nil) break; /* deactivated by unforward */
        if(((long)c->car & FL_INT_BIT) != 0) {
            if(tcb->state == FL_FINISHING) {
                int id = RPTR_ID(c->car);
                unref(&fl_tcbs[ id - 1 ], RPTR(c->car));
            } else {
                FL_VAL ev = mklist(tcb, MKINT(SEND_DROPREF),
                                   mklist(tcb, c->car, fl_nil));
                fl_add_event(tcb, ev);
            }
        } else {
            close_stream(tcb, c->car);
            unref(tcb, c->car);
        }
        break;
    case FL_REF_TAG:
        if(c->car == fl_nil) break; /* deactivated by unforward */
        if(tcb->state == FL_FINISHING) {
            int id = RPTR_ID(c->car);
            unref(&fl_tcbs[ id - 1 ], RPTR(c->car));
        } else {
            tl = mklist(tcb, MKINT(SEND_DROPREF),
                        mklist(tcb, c->car, fl_nil));
            fl_add_event(tcb, tl);
        }
        break;
    default:
        unref(tcb, c->car);
        unref(tcb, c->cdr);
        break;
    }
    /* XXX make this tail-recursive */
    c->tag = 0;
    c->car = (FL_VAL)FL_BROKEN_HEART;
    c->cdr = tcb->freelist;
    tcb->freelist = c;
    --tcb->used;
}

void fl_error(FL_TCB *tcb, FL_VAL err)
{
    fl_write(tcb, stderr, err);
    fputc('\n', stderr);
    terminate(1);
}

void fl_fail(FL_TCB *tcb)
{
    fprintf(stderr, "<%d> %ld: %s failed\n", tcb->ordinal, tcb->ticks, tcb->goal->info);
    terminate(1);
}

static char **next_string(char *p)
{
    // add length- and final 0-byte
    int wsize = sizeof(long);
    return (char **)(((long)p + *((unsigned char *)p) + wsize + 1) & ~(wsize - 1));
}

FL_VAL fl_intern(char *np)
{
    pthread_mutex_lock(&intern_mutex);
    char *p = strings; 
    while(p != NULL) {
        if(*p == *np && !strncmp(np + 1, p + 1, *p)) {
            pthread_mutex_unlock(&intern_mutex);
            return p;
        }
        p = *next_string(p);
    }
    *next_string(np) = strings;
    strings = np;
    pthread_mutex_unlock(&intern_mutex);
    return np;
}

FL_VAL fl_mkstring(char *str, int len)
{
    char *buf = malloc(len + 2 + 2 + sizeof(char *));
    char *sp = buf + 2;
    *sp = len;
    memcpy(sp + 1, str, len);
    sp[ len + 2 ] = '\0';
    int wsize = sizeof(long);
    *((char **)(((long)sp + len + wsize + 1) & ~(wsize - 1))) = NULL;
    FL_VAL s = fl_intern(sp);
    if(s != sp) free(buf);
    return s;
}

void fl_send_to_port(FL_TCB *tcb, FL_VAL port, FL_VAL x)
{
    if(BITS(port) != 0 || TAG(port) != FL_PORT_TAG)
        fl_rt_error(tcb, port, FL_NOT_A_PORT);
    if(((long)CAR(port) & FL_INT_BIT) != 0) {
        FL_VAL ev = mklist(tcb, (FL_VAL)((long)(SEND_SENDPORT << 1) | 1),
                           mklist(tcb, CAR(port), x));
        fl_add_event(tcb, ev);
        return;
    }
    FL_VAL head = CAR(port);
    FL_VAL st = fl_chase_tail(head);
    FL_VAL e = fl_alloc_cell(tcb, fl_nil, fl_nil);
    ((FL_CELL *)e)->tag = FL_VAR_TAG;
    ((FL_CELL *)e)->car = e;
    ((FL_CELL *)port)->car = addref(e);
    FL_VAL n = fl_alloc_cell(tcb, x, e);
    ((FL_CELL *)n)->tag = FL_LIST_TAG;
    fl_assign(tcb, n, st);
    unref(tcb, head);
}

static unsigned char *forward_rec(FL_TCB *tcb, FL_VAL x, int dest, 
    unsigned char *ptr, unsigned char *limit)
{
    for(;;) {
        x = deref(x);
        assert(ptr < limit);    /* XXX check limit and forward var */
        if(BITS(x) != 0) {
            *((FL_VAL *)ptr) = x;
            return ptr + sizeof(FL_VAL);
        }
        FL_VAL ev;
        switch(TAG(x)) {
        case FL_FLOAT_TAG:
            *((unsigned long *)ptr) = FL_FLOAT_TAG;
            ptr += sizeof(long);
#if !defined(__LP64__) && !defined(_LP64)
            if(((long)ptr & 8) != 0) ptr += sizeof(long);
#endif
            *((double *)ptr) = *((double *)&CAR(x));
            return ptr + sizeof(double);
        case FL_PORT_TAG:
            *((unsigned long *)ptr) = FL_PORT_TAG;
            ptr += sizeof(long);
            if(((long)CAR(x) & FL_INT_BIT) != 0) {
                *((FL_VAL *)ptr) = CAR(x);
                if(RPTR_ID(CAR(x)) != dest) {
                    ev = mklist(tcb, MKINT(SEND_ADDREF), 
                                mklist(tcb, CAR(x), fl_nil));
                    fl_add_event(tcb, ev);
                } else addref(RPTR(CAR(x)));
            } else *((FL_VAL *)ptr) = mkrptr(tcb, addref(x));
            return ptr + sizeof(FL_VAL);
        case FL_VAR_TAG:
            if(dest == tcb->ordinal) {
                *((unsigned long *)ptr) = FL_VAR_TAG;
                ptr += sizeof(long);
                *((FL_VAL *)ptr) = addref(x);
            } else {
                *((unsigned long *)ptr) = FL_REF_TAG;
                ptr += sizeof(long);
                FL_VAL rptr = mkrptr(tcb, addref(x));
                *((FL_VAL *)ptr) = rptr;
            }
            return ptr + sizeof(FL_VAL);
        case FL_REF_TAG:
            *((unsigned long *)ptr) = FL_REF_TAG;
            ptr += sizeof(long);
            *((FL_VAL *)ptr) = CAR(x);
            if(RPTR_ID(CAR(x)) != dest) {
                ev = mklist(tcb, MKINT(SEND_ADDREF), 
                            mklist(tcb, CAR(x), fl_nil));
                fl_add_event(tcb, ev);
            }
            return ptr + sizeof(FL_VAL);
        case FL_LIST_TAG:
            *((unsigned long *)ptr) = FL_LIST_TAG;
            ptr += sizeof(long);
            ptr = forward_rec(tcb, CAR(x), dest, ptr, limit);
            x = CDR(x);
            break;
        case FL_MODULE_TAG:
            *((unsigned long *)ptr) = FL_MODULE_TAG;
            ptr += sizeof(long);
            *((FL_MODULE **)ptr) = (FL_MODULE *)CAR(x);
            return ptr + sizeof(FL_MODULE *);
        default:
            if((TAG(x) & FL_TUPLE_BIT) != 0) {
                *((unsigned long *)ptr) = TAG(x);
                ptr += sizeof(long);
                ptr = forward_rec(tcb, CAR(x), dest, ptr, limit);
                x = CDR(x);
            } else fl_rt_error(tcb, x, FL_CANT_FORWARD);
            break;
        }
    }
}

int fl_forward(FL_TCB *tcb, int id, FL_VAL x)
{
    FL_TCB *dest = &fl_tcbs[ id - 1 ];
    while(!__sync_bool_compare_and_swap(&dest->port_lock, 0, 1));
    if(dest->port_size) {
        __sync_synchronize();
        dest->port_lock = 0;
        return 0;
    }
#ifdef LOGGING
    if(tcb->logging) {
        pthread_mutex_lock(&log_mutex);
        fprintf(logfile, "<%d> %ld: %s send -> <%d>: ", tcb->ordinal, tcb->ticks, 
            tcb->goal->info, id);
        fl_write(tcb, logfile, x);
        fputc('\n', logfile);
        pthread_mutex_unlock(&log_mutex);
    }
#endif
    unsigned char *p = forward_rec(tcb, x, id, dest->port, dest->port + PORT_SIZE);
#ifdef XXLOGGING
    if(tcb->logging) {
        pthread_mutex_lock(&log_mutex);
        for(unsigned char *pp = dest->port; pp < p; pp += sizeof(long))
            fprintf(logfile, "<%d>    0x%08lx\n", tcb->ordinal,
                *((unsigned long *)pp));
        pthread_mutex_unlock(&log_mutex);
    }        
#endif
    dest->port_size = p - dest->port;
    dest->port_source = tcb->ordinal;
    idle_threads = 0;
    __sync_synchronize();
    dest->port_lock = 0;
    return 1;
}

static unsigned char *unforward_rec(FL_TCB *tcb, unsigned char *ptr, FL_VAL *result)
{
    for(;;) {
        FL_VAL x = *((FL_VAL *)ptr);
        ptr += sizeof(FL_VAL);
        if(BITS(x) != 0) {
            *result = x;
            return ptr;
        }
        FL_VAL y = fl_alloc_cell(tcb, fl_nil, fl_nil);
        ((FL_CELL *)y)->tag = (unsigned long)x;
        *result = addref(y);
        switch((unsigned long)x) {
        case FL_PORT_TAG:
            CAR(y) = *((FL_VAL *)ptr);
            ptr += sizeof(FL_VAL);
            if(RPTR_ID(CAR(y)) == tcb->ordinal) {
                *result = RPTR(CAR(y));
                CAR(y) = fl_nil;    /* deactivate */
                fl_release(tcb, y);
            }
            return ptr;
        case FL_VAR_TAG:    /* locally sent to self */
            fl_release(tcb, y);
            *result = *((FL_VAL *)ptr);
            ptr += sizeof(FL_VAL);
            return ptr;
        case FL_REF_TAG:
            CAR(y) = *((FL_VAL *)ptr);
            ptr += sizeof(FL_VAL);
            if(RPTR_ID(CAR(y)) == tcb->ordinal) {
                /* rptr in REF originated from this thread: */
                *result = addref(RPTR(CAR(y)));
                /* change tag, or logging may crash as RPTR is invalid: */
                ((FL_CELL *)y)->tag = FL_LIST_TAG;
                CAR(y) = fl_nil; /* deactivate */
                fl_release(tcb, y);
            }
            return ptr;
        case FL_FLOAT_TAG:
#if !defined(__LP64__) && !defined(_LP64)
            if(((long)ptr & 8) != 0) ptr += sizeof(long);
            if(((long)y & 4) == 0) {
                fl_release(tcb, y);
                y = fl_alloc_float(tcb, *((double *)ptr));
                *result = addref(y);
                return ptr + sizeof(double);
            }
#endif
            *((double *)&CAR(y)) = *((double *)ptr);
            return ptr + sizeof(double);
        case FL_LIST_TAG:
            ptr = unforward_rec(tcb, ptr, &CAR(y));
            result = &CDR(y);
            break;
        case FL_MODULE_TAG:
            *((FL_MODULE **)&CAR(x)) = *((FL_VAL *)ptr);
            return ptr + sizeof(FL_MODULE *);
        default:
            if(((unsigned long)x & FL_TUPLE_BIT) != 0) {
                ptr = unforward_rec(tcb, ptr, &CAR(y));
                result = &CDR(y);
                break;
            } else fl_rt_error(tcb, fl_nil, FL_CANT_UNFORWARD);
        }
    }
}

FL_VAL fl_unforward(FL_TCB *tcb)
{
    while(!__sync_bool_compare_and_swap(&tcb->port_lock, 0, 1));
    FL_VAL x;
    unsigned char *p = tcb->port + tcb->port_size;
    unforward_rec(tcb, tcb->port, &x);
    tcb->port_size = 0;
    int src = tcb->port_source;
    __sync_synchronize();
    tcb->port_lock = 0;
#ifdef LOGGING
    if(tcb->logging) {
        pthread_mutex_lock(&log_mutex);
        fprintf(logfile, "<%d> %ld: %s receive <- <%d>: ",
             tcb->ordinal, tcb->ticks, tcb->goal->info, src);
        fl_write(tcb, logfile, x);
        fputc('\n', logfile);
#ifdef XXLOGGING
        for(unsigned char *pp = tcb->port; pp < p; pp += sizeof(long))
            fprintf(logfile, "<%d>    0x%08lx\n", tcb->ordinal,
                *((unsigned long *)pp));
#endif
        pthread_mutex_unlock(&log_mutex);
    }
#endif
    return x;
}

void fl_init_module(FL_MODULE *mod)
{
    mod->next = last_module;
    last_module = mod;
    int i = 0;
    char *np = mod->names;
    while(np != NULL) {
        char *np2 = *next_string(np);
        mod->atoms[ i++ ] = fl_intern(np);
        np = np2;
    }
    for(FL_VAL **rap = mod->ratoms; *rap != NULL; ++rap) {
        FL_VAL *ap = *rap;
        *ap = mod->atoms[ (long)(*ap) ];
    }
    mod->name = mod->atoms[ (long)mod->name ];
    for(i = 0; i < mod->npdefs; ++i)
        mod->pdefs[ i ].name = mod->atoms[ (long)mod->pdefs[ i ].name ];
}

static void init_thread(FL_TCB *tcb, long heapsz, int maxgoals)
{
    tcb->goal = NULL;
    tcb->sstack = (FL_VAL *)malloc(MAX_SSTACK * sizeof(FL_VAL) * 2);
    tcb->sstackp = tcb->sstack;
    tcb->tailcalls = 0;
    tcb->timeslice = timeslice;
    tcb->ticks = 0;
    tcb->stackbase = NULL;  /* set on thread-start */
    long cells = heapsz / sizeof(FL_CELL);
    tcb->heap = (FL_CELL *)malloc(heapsz);
    tcb->heapend = tcb->heap + cells;
    tcb->heap[ 0 ].car = (FL_VAL)FL_BROKEN_HEART;
    tcb->heap[ 0 ].tag = 0;
    tcb->heap[ 0 ].cdr = NULL;
    for(int i = 1; i < cells; ++i) {
        tcb->heap[ i ].tag = 0;
        tcb->heap[ i ].car = (FL_VAL)FL_BROKEN_HEART;
        tcb->heap[ i ].cdr = (FL_VAL)(&tcb->heap[ i - 1 ]);
    }
    tcb->freelist = &tcb->heap[ cells - 1 ];
    tcb->suspended = 0;
    tcb->active = 0;
    tcb->goal_buffer = (FL_GOAL *)malloc(maxgoals * sizeof(FL_GOAL));
    tcb->max_goals = maxgoals;
    tcb->queue = (FL_GOAL **)malloc(maxgoals * sizeof(FL_GOAL *));
    tcb->qstart = tcb->qend = tcb->queue;
    tcb->goal_buffer[ 0 ].addr = NULL;
    for(int i = 1; i < maxgoals; ++i)
        tcb->goal_buffer[ i ].addr = (void *)(&tcb->goal_buffer[ i - 1 ]);
    tcb->freegoals = &tcb->goal_buffer[ maxgoals - 1 ];
    tcb->delay = 0;
    tcb->events = fl_nil;
    tcb->listening = fl_nil;
    tcb->state = FL_BUSY;
    tcb->port_size = 0;
    tcb->port = malloc(PORT_SIZE);
    tcb->port_lock = 0;
    tcb->defer = 0;
    tcb->idle = fl_alloc_cell(tcb, fl_nil, fl_nil);
    ((FL_CELL *)tcb->idle)->tag = FL_VAR_TAG | 1;
    ((FL_CELL *)tcb->idle)->car = tcb->idle;
    if(fl_create_event_queue(tcb) == -1) {
        fprintf(stderr, "\n*** ERROR: <%d> unable to create event queue: %s\n",
            tcb->ordinal, strerror(errno));
        exit(1);
    }
}

static void init_goal(FL_TCB *tcb, FL_PDEF *pdef)
{
#ifdef LOGGING
    if(tcb->logging) 
        fl_logfmt(tcb, "initializing goal %s/%ld @ %p\n", (char *)pdef->name + 1, 
            pdef->arity, pdef->addr);
#endif
    FL_GOAL *g = tcb->freegoals;
    tcb->freegoals = (FL_GOAL *)g->addr;
    g->addr = pdef->addr;
    ++tcb->goals;
    if(tcb->goal != NULL) enqueue(tcb, g);
    else tcb->goal = g;
}

void fl_init_goal(FL_PDEF *pdef)
{
    init_goal(&fl_tcbs[ 0 ], pdef);
}

void fl_resolve_pdef(FL_TCB *tcb, long arity, FL_VAL mod, FL_VAL name, void *cache, 
    FL_GOAL *ng)
{
#ifdef XXLOGGING
    if(tcb->logging)
        fl_logfmt(tcb, "resolve %s/%ld, cache=%p\n", (char *)name + 1, arity, cache);
#endif
    FL_VAL lst;
    int many = 0;
    int ar = arity;
    if(arity > 4) {
        lst = ng->args[ 3 ];
        ar = 3;
        many = 1;
    } else lst = fl_nil;
    for(int i = ar - 1; i >= 0; --i) {
        FL_VAL lst2 = mklist(tcb, ng->args[ i ], lst);
        unref(tcb, ng->args[ i ]);
        lst = lst2;
    }
    if(many) unref(tcb, ng->args[ 3 ]);
    lst = mklist(tcb, mod, mklist(tcb, mktuple(tcb, name, arity, lst), fl_nil));
    ng->args[ 0 ] = addref(mktuple(tcb, fl_colon, 2, lst));
    ng->args[ 1 ] = (FL_VAL)((long)cache | 1);
}

static int type_order(FL_VAL x)
{
    if(((long)x & FL_INT_BIT) != 0) return 1;
    if(((long)x & FL_ATOM_BIT) != 0) return 3;
    if((TAG(x) & FL_TUPLE_BIT) != 0) return 5;
    switch(TAG(x)) {
    case FL_VAR_TAG: return 0;
    case FL_FLOAT_TAG: return 2;
    case FL_LIST_TAG: return 4;
    case FL_PORT_TAG: return 6;
    case FL_MODULE_TAG: return 8;
    }
    return 9; /* ??? */
}

static int signum(long n)
{   
    if(n < 0) return -1;
    return n > 0 ? 1 : 0; 
}

int fl_ordering(FL_VAL x, FL_VAL y)
{
    x = deref(x);
    y = deref(y);
    int o1 = type_order(x), o2 = type_order(y);
    if(o1 < o2) return -1;
    if(o1 > o2) return 1;
    switch(o1) {
    case 2: /* float */ {
        double f1 = *((double *)&CAR(y));
        double f2 = *((double *)&CAR(y));
        return signum((int)(f1 - f2)); }
    case 3: /* string */ {
        int l1 = *((unsigned char *)x);
        int l2 = *((unsigned char *)y);
        char *p1 = (char *)x + 1;
        char *p2 = (char *)y + 1;
        for(int i = 0; i < l1; ++i) {
            if(i >= l2) return -1;
            if(*p1 < *p2) return -1;
            if(*p1 > *p2) return 1;
            ++p1;
            ++p2;
        }
        return 0; }
    case 4: /* list */ {
        int o = fl_ordering(CAR(x), CAR(y));
        if(o < 0) return -1;
        if(o > 0) return 1;
        return fl_ordering(CDR(x), CDR(y)); }
    case 5: /* tuple */ {
        int l1 = (((FL_CELL *)x)->tag & ~FL_TUPLE_BIT) >> 24;
        int l2 = (((FL_CELL *)y)->tag & ~FL_TUPLE_BIT) >> 24;
        if(l1 < l2) return -1;
        if(l1 > l2) return 1;
        int o = fl_ordering(CAR(x), CAR(y));
        return o == 0 ? fl_ordering(CDR(x), CDR(y)) : 0; }
    case 8: /* module */
        return (int)(CAR(x) - CAR(y));
    default:
        return signum((long)x - (long)y);
    }
}

FL_VAL fl_unify_rec(FL_TCB *tcb, FL_VAL x, FL_VAL y)
{
    for(;;) {
        if(x == y) return fl_true;
        if(((long)x & FL_BITS_MASK) == 0) {
            if(TAG(x) == FL_VAR_TAG) {
                if(((FL_CELL *)x)->car != x) {
                    x = ((FL_CELL *)x)->car;
                    continue;
                }
            }
        }
        if(((long)y & FL_BITS_MASK) == 0) {
            if(TAG(y) == FL_VAR_TAG) {
                if(((FL_CELL *)y)->car != y) {
                    y = ((FL_CELL *)y)->car;
                    continue;
                }
            }
        }
        if(((long)x & FL_BITS_MASK) == 0) {
            if((TAG(x) == FL_VAR_TAG && ((FL_CELL *)x)->car == x) || 
                TAG(x) == FL_REF_TAG) {
                fl_assign(tcb, y, x);
                return fl_true;
            }
        } 
        if(((long)y & FL_BITS_MASK) != 0) return fl_false;
        if((TAG(y) == FL_VAR_TAG && ((FL_CELL *)y)->car == y) || 
            TAG(y) == FL_REF_TAG) {
            fl_assign(tcb, x, y);
            return fl_true;
        }
        if(((long)x & FL_BITS_MASK) != 0) return fl_false;
        long tx = TAG(x);
        if(tx != TAG(y)) return fl_false;
        if(tx == FL_FLOAT_TAG)
            return *((double *)&CAR(x)) == *((double *)&CAR(y)) ? fl_true : fl_false;
        if(tx == FL_PORT_TAG) return fl_false;
        if(fl_unify_rec(tcb, ((FL_CELL *)x)->car, ((FL_CELL *)y)->car) != fl_true)
            return fl_false;
        x = ((FL_CELL *)x)->cdr;
        y = ((FL_CELL *)y)->cdr;
    }
}

void fl_unify_result(FL_TCB *tcb, FL_VAL val, FL_VAL dest)
{
    FL_VAL r = fl_unify_rec(tcb, addref(val), dest);
    unref(tcb, val);
    if(r == fl_false)
        fl_rt_error2(tcb, val, dest, FL_CANT_UNIFY);
}

static int check_heap(FL_TCB *tcb)
{
    if(!tcb->used) return 1;
    fprintf(stderr, "\n*** ERROR: <%d> memory leak of %ld cells:\n", tcb->ordinal, 
        tcb->used);
    for(FL_CELL *p = tcb->heap; p < tcb->heapend; ++p) {
        if((p->tag & FL_COUNT_MASK) != 0) {
            fprintf(stderr, "<%d>   %p 0x%08x ", tcb->ordinal, p, (int)p->tag);
            fl_write(tcb, stderr, (FL_VAL)p);
            fputc('\n', stderr);
        }
    }
    return 0;
}

static void finalize_heaps()
{
    int i;
#ifdef LOGGING
    if(fl_tcbs[ 0 ].logging) 
        fl_logfmt(&fl_tcbs[ 0 ], "waiting for threads\n");
#endif
    for(i = 1; i < threads; ++i) 
        pthread_join(fl_tcbs[ i ].thread, NULL);
    for(i = 0; i < threads; ++i) {
        unref(&fl_tcbs[ i ], fl_tcbs[ i ].listening);
        unref(&fl_tcbs[ i ], fl_tcbs[ i ].idle);
    }
    int failed = 0;
    for(i = 0; i < threads; ++i)
        failed = !check_heap(&fl_tcbs[ i ]) || failed;
    if(failed) terminate(1);
}

void fl_yield(FL_TCB *tcb)
{
    ++tcb->ticks;
    if((tcb->ticks % stat_ticks) == 0) {
        if((stats & (1 << (tcb->ordinal - 1))) != 0) 
            statistics(tcb);
        if((heapstats & (1 << (tcb->ordinal - 1))) != 0) 
            heap_statistics(tcb);
    }
listen:
    if((idle_threads & THREAD_BIT(tcb)) == 0 && tcb->state == FL_IDLE) {
        tcb->state = FL_BUSY;
#ifdef XLOGGING
        if(tcb->logging) fl_logfmt(tcb, "IDLE -> BUSY (forced)\n");
#endif
    }
    if(tcb->port_size > 0) {
        FL_VAL ev = fl_unforward(tcb);
        fl_add_event(tcb, ev);
        unref(tcb, ev); /* unforward adds one implicit ref */
        if(tcb->state != FL_BUSY) {
            assert(tcb->state == FL_IDLE);
            tcb->state = FL_BUSY;
#ifdef XLOGGING
            if(tcb->logging) 
                fl_logfmt(tcb, "IDLE -> BUSY (message)\n");
#endif
            idle_threads = 0;
        }
    }
    if(tcb->listening != fl_nil) {
        int ms = tcb->delay >> 8;
        if(tcb->qstart == tcb->qend && threads == 1) {
#ifdef XLOGGING
            if(tcb->logging) fl_logfmt(tcb, "waiting for event\n");
#endif
            ms = -1;
        }
#ifdef XLOGGING
        if(tcb->logging) {
            pthread_mutex_lock(&log_mutex);
            fprintf(logfile, "<%d> %ld: %s listen: ", tcb->ordinal, tcb->ticks,
                tcb->goal->info);
            fl_write(tcb, logfile, tcb->listening);
            fputc('\n', logfile);
            pthread_mutex_unlock(&log_mutex);
        }
#endif        
        int r = fl_wait_for_events(tcb, ms);
        if(r == -1) fl_rt_error(tcb, fl_nil, FL_IO_ERROR);
        else if(r > 0) tcb->delay = 0;
        else if(ms != -1) {
            tcb->delay += FUDGE;
            if(tcb->delay > MAX_DELAY) tcb->delay = MAX_DELAY;
        }
    }
    if(tcb->qstart == tcb->qend) {
        if(CDR(tcb->idle) != fl_nil) {
            fl_assign(tcb, fl_nil, tcb->idle);
            unref(tcb, tcb->idle);
            tcb->idle = fl_alloc_cell(tcb, fl_nil, fl_nil);
            ((FL_CELL *)tcb->idle)->tag = FL_VAR_TAG | 1;
            ((FL_CELL *)tcb->idle)->car = tcb->idle;
            goto next;
        }
        if(tcb->listening != fl_nil) goto listen;
        if(tcb->state == FL_BUSY) {
            long n = __sync_or_and_fetch(&idle_threads, THREAD_BIT(tcb));
            tcb->state = FL_IDLE;
            if(tcb->logging) fl_logfmt(tcb, "BUSY -> IDLE (0x%lx)\n", n);
            /* give other threads a chance to pick up any pending messages
               in case they are sleeping: */
            tcb->delay = MAX_DELAY * 2;
        } else if(idle_threads == (1 << threads) - 1) {
            if(tcb->state == FL_FINISHING) {
                if(tcb->suspended != 0) {
                    fprintf(stderr, "<%d> deadlock with %ld suspended goal(s)\n", 
                        tcb->ordinal, tcb->suspended);
                    terminate(1);
                }
#ifdef LOGGING
                if(tcb->logging) fl_logfmt(tcb, "terminates\n");
#endif
                if(tcb->ordinal == 1) finalize_heaps();
#ifdef LOGGING
                if(tcb->logging && tcb->ordinal == 1) 
                    fl_logfmt(tcb, "exit\n");
#endif
                fflush(logfile);
                pthread_exit(NULL);
                /* never reached */
            } else if(tcb->state == FL_IDLE) {
#ifdef LOGGING
                if(tcb->logging) fl_logfmt(tcb, "shutting down\n");
#endif
                close_stream(tcb, tcb->events);
#ifdef XLOGGING
                if(tcb->logging) 
                    fl_logfmt(tcb, "IDLE -> FINISHING\n");
#endif
                tcb->state = FL_FINISHING;
                unref(tcb, tcb->events);
            }
        }
        if(usleep(tcb->delay >> 8) == -1) tcb->delay = 0; /* likely EINTR */
        else {
            tcb->delay += FUDGE;
            if(tcb->delay > MAX_DELAY) tcb->delay = MAX_DELAY;
        }
        goto listen;
    }
next:
    tcb->goal = dequeue(tcb);
    fl_enter(tcb);
}

static void *enter_thread(void *tcb0) 
{
    FL_TCB *tcb = (FL_TCB *)tcb0;
    tcb->stackbase = alloca(MAX_ENV * sizeof(FL_VAL));
    fl_enter(tcb);
    return NULL;    /* never reached */
}

static void build_topology(void)
{
    int h = round(sqrt(threads));
    int w = h + (threads - h * h) / h;
    for(int i = 0; i < threads; ++i) {
        int right = (i + 1) % threads;
        fl_tcbs[ i ].peers[ -FL_FWD - 1 ] = right + 1;
        fl_tcbs[ right ].peers[ -FL_BWD - 1 ] = i + 1;
    }
    for(int r = 0; r < h; ++r) {
        for(int c = 0; c < w; ++c) {
            int here = (r * w) + c;
            int below = ((r + 1) % h) * w + c;
            fl_tcbs[ here ].peers[ -FL_SOUTH - 1 ] = below + 1;
            fl_tcbs[ below ].peers[ -FL_NORTH - 1 ] = here + 1;
            int right = (r * w) + (c + 1) % w;
            fl_tcbs[ here ].peers[ -FL_EAST - 1 ] = right + 1;
            fl_tcbs[ right ].peers[ -FL_WEST - 1 ] = here + 1;
        }
    }
}

FL_MODULE *fl_find_module(FL_TCB *tcb, FL_VAL name)
{
    for(FL_MODULE *m = last_module; m != NULL; m = m->next) {
        if(m->name == name) return m;
    }
    return NULL;
}

void fl_start(int threads, long heapsz, int maxgoals)
{
    fl_nil = fl_mkstring("[]", 2);
    fl_colon = fl_mkstring(":", 1);
    fl_true = fl_mkstring("true", 4);
    fl_false = fl_mkstring("false", 5);
    sys_name = fl_mkstring("sys", 3);
    for(int i = 0; i < threads; ++i) {
        fl_tcbs[ i ].ordinal = i + 1;
        fl_tcbs[ i ].logging = (logging & (1 << i)) != 0;
        init_thread(&fl_tcbs[ i ], heapsz, maxgoals);
    }
    build_topology();
    fl__init();
    FL_MODULE *sysmod = fl_find_module(&fl_tcbs[ 0 ], sys_name);
    assert(threads == 1 || sysmod != NULL);
    pthread_attr_t attr;
    pthread_attr_init(&attr);
    pthread_attr_setdetachstate(&attr, PTHREAD_CREATE_JOINABLE);
    for(int i = 1; i < threads; ++i) {
        init_goal(&fl_tcbs[ i ], sysmod->pdefs);
        pthread_create(&fl_tcbs[ i ].thread, &attr, enter_thread, (void *)&fl_tcbs[ i ]);
    }
    enter_thread(&fl_tcbs[ 0 ]);
}

static void usage(char *prg, int code)
{
    fprintf(stderr, "usage: %s [+HELP] [+LOG <mask>] [+HEAP <heapsize>] [+GOALS <maxgoals>] [+TIMESLICE <ticks>] [+LOGFILE <filename>] [+THREADS <number>] [+STATS] [+STAT_TICKS <number>] [+HEAPSTATS] ... [--] ...\n", prg);
    exit(code);
}

int main(int argc, char *argv[])
{
    int heap_size = HEAP_SIZE;
    int max_goals = MAX_GOALS;
    logfile = stderr;
    int args = 0;
    int i;
    threads = 1;
    idle_threads = 0;
    struct timeval tv;
    gettimeofday(&tv, NULL);
    fl_set_random_seed((unsigned char *)&tv, sizeof(struct timeval));
    for(i = 1; i < argc; ++i) {
        if(!strcmp("--", argv[ i ])) break;
        if(*argv[ i ] == '+') {
            if(!strcmp("+LOG", argv[ i ]))
                logging = atoi(argv[ ++i ]);
            else if(!strcmp("+STATS", argv[ i ]))
                stats = atoi(argv[ ++i ]); 
            else if(!strcmp("+STAT_TICKS", argv[ i ]))
                stat_ticks = atoi(argv[ ++i ]); 
            else if(!strcmp("+HEAPSTATS", argv[ i ]))
                heapstats = atoi(argv[ ++i ]); 
            else if(!strcmp("+HEAP", argv[ i ]))
                heap_size = atoi(argv[ ++i ]);
            else if(!strcmp("+GOALS", argv[ i ]))
                max_goals = atoi(argv[ ++i ]);
            else if(!strcmp("+THREADS", argv[ i ]))
                threads = atoi(argv[ ++i ]);
            else if(!strcmp("+TIMESLICE", argv[ i ]))
                timeslice = atoi(argv[ ++i ]);
            else if(!strcmp("+LOGFILE", argv[ i ])) {
                FILE *fp;
                char *fname = argv[ ++i ];
                if((fp = fopen(fname, "a")) == NULL) {
                    fprintf(stderr, "can not open log file %s\n", fname);
                    exit(1);
                }
                logfile = fp;
            }
            else if(!strcmp("+HELP", argv[ i ]))
                usage(argv[ 0 ], 0);
            else usage(argv[ 0 ], 1);
        } else fl_argv[ args++ ] = argv[ i ];
    }
    assert(threads >= 1 && threads <= MAX_THREADS);
    while(i < argc) fl_argv[ args++ ] = argv[ i++ ];
    fl_argv[ args ] = NULL;
    fl_start(threads, heap_size, max_goals);
}
