/* FLENG - runtime system - library */

#include "fleng.h"
#include "fleng-util.h"
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <assert.h>
#include <limits.h>
#include <errno.h>
#include <math.h>
#include <fcntl.h>
#include <unistd.h>
#include <stdint.h>
#include <sys/time.h>

#define INIT_CBUF   10000
#define MAX_ARGS    256

void fl_writeln_2(FL_TCB *tcb, FL_VAL x, FL_VAL done)
{
    fl_write(tcb, stdout, x);
    fputc('\n', stdout);
    fflush(stdout);
    fl_assign(tcb, fl_nil, done);
}

void fl_find_module_2(FL_TCB *tcb, FL_VAL name, FL_VAL var)
{
    FL_MODULE *m = fl_find_module(tcb, name);
    if(m != NULL) {
        FL_VAL mod = (FL_VAL)((long)m | 1);
        fl_assign(tcb, mod, var);
        return;
    }
    fl_rt_error(tcb, name, FL_NO_MODULE);
}

void fl_find_pdef_4(FL_TCB *tcb, FL_VAL module, FL_VAL name, FL_VAL arity, 
    FL_VAL var)
{
    FL_MODULE *mod = (FL_MODULE *)((long)module & ~1);
    int a = INT(arity);
    for(int i = 0; i < mod->npdefs; ++i) {
        if(mod->pdefs[ i ].name == name && mod->pdefs[ i ].arity == a) {
            fl_assign(tcb, (FL_VAL)((long)(mod->pdefs[ i ].addr) | 1), var);
            return;
        }
    }
    fl_rt_error(tcb, name, FL_NO_PDEF);
}

void fl_arity_2(void *tcb, FL_VAL x, FL_VAL var)
{
    long n = (TAG(x) >> 24) & 0x7f;
    fl_assign(tcb, (FL_VAL)(((long)n << 1) | 1), var);
}

void fl_uncons_3(FL_TCB *tcb, FL_VAL lst, FL_VAL var1, FL_VAL var2)
{
    fl_assign(tcb, CAR(lst), var1);
    fl_assign(tcb, CDR(lst), var2);
}

void fl_set_cache_2(FL_TCB *tcb, FL_VAL cache, FL_VAL pdef)
{
    *((long *)((long)cache & ~1)) = (long)pdef & ~1;
}

void fl_string_length_2(FL_TCB *tcb, FL_VAL x, FL_VAL var)
{
    unsigned char *p = (char *)x;
    int len = 0, n = *(p++);
    unsigned char *pe = p + n;
    while(p < pe) {
        unsigned char c = *p;
        if((c & 0xf8) == 0xf0) p += 4;
        else if((c & 0xf0) == 0xe0) p += 3;
        else if((c & 0xe0) == 0xc0) p += 2;
        else ++p;
        ++len;
    }
    fl_assign(tcb, (FL_VAL)(((long)len << 1) | 1), var);
}

void fl_bytes_length_2(FL_TCB *tcb, FL_VAL x, FL_VAL var)
{
    fl_assign(tcb, CDR(CAR(x)), var);
}

void fl_tuple_length_2(FL_TCB *tcb, FL_VAL x, FL_VAL var)
{
    fl_assign(tcb, MKINT(((TAG(deref(x)) >> 24) & 0x7f) + 1), var);
}

void fl_command_line_1(FL_TCB *tcb, FL_VAL v)
{
    FL_VAL lst = fl_nil;
    char **p = fl_argv;
    while(*p != NULL) ++p;
    while(p-- > fl_argv) {
        lst = fl_alloc_cell(tcb, fl_mkstring(*p, strlen(*p)), lst);
        ((FL_CELL *)lst)->tag = FL_LIST_TAG;
    }
    fl_unify_result(tcb, lst, v);
}

static void str_to_int(FL_TCB *tcb, char *str, int base, FL_VAL var)
{
    char *end;
    long n = strtol(str, &end, base);
    if(*end != '\0') goto fail;
    if((n == LONG_MIN || n == LONG_MAX) && errno == ERANGE) goto fail;
    if(n == 0 && errno == EINVAL) goto fail;
    fl_assign(tcb, (FL_VAL)((n << 1) | 1), var);
    return;
fail:
    fl_rt_error(tcb, str, FL_CANT_CONVERT);
}

void fl_string_to_integer_3(FL_TCB *tcb, FL_VAL str, FL_VAL base, FL_VAL var)
{
    if(((long)base & 1) != 1) fl_rt_error(tcb, base, FL_NOT_AN_INT);
    int b = INT(base);
    if(BITS(str) != 2) fl_rt_error(tcb, str, FL_NOT_AN_ATOM);
    str_to_int(tcb, str + 1, b, var);
}

void fl_string_to_integer_2(FL_TCB *tcb, FL_VAL str, FL_VAL var)
{
    fl_string_to_integer_3(tcb, str, (FL_VAL)((10 << 1) | 1), var);
}

static void str_to_float(FL_TCB *tcb, char *str, FL_VAL var) 
{
    char *end;
    double n = strtod(str, &end);
    if(*end != '\0') goto fail;
    if((n == HUGE_VAL || n == -HUGE_VAL || n == 0) && errno == ERANGE)
        goto fail;
    FL_VAL c = fl_alloc_cell(tcb, (FL_VAL)1, (FL_VAL)1);
    ((FL_CELL *)c)->tag = FL_FLOAT_TAG;
    *((double *)&((FL_CELL *)c)->car) = n;
    fl_assign(tcb, c, var);
    return;
fail:
    fl_rt_error(tcb, str, FL_CANT_CONVERT);
}

void fl_string_to_float_2(FL_TCB *tcb, FL_VAL str, FL_VAL var)
{
    char *end;
    if(BITS(str) != 2) fl_rt_error(tcb, str, FL_NOT_AN_ATOM);
    str_to_float(tcb, (char *)str + 1, var);
}

/* Branchless UTF-8 decoder
 * https://github.com/skeeto/branchless-utf8/blob/master/utf8.h
 * This is free and unencumbered software released into the public domain. */
static unsigned char *utf8_decode(unsigned char *buf, uint32_t *c)
{
    static const char lengths[] = {
        1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1,
        0, 0, 0, 0, 0, 0, 0, 0, 2, 2, 2, 2, 3, 3, 4, 0
    };
    static const int masks[]  = {0x00, 0x7f, 0x1f, 0x0f, 0x07};
    static const uint32_t mins[] = {4194304, 0, 128, 2048, 65536};
    static const int shiftc[] = {0, 18, 12, 6, 0};
    static const int shifte[] = {0, 6, 4, 2, 0};
    unsigned char *s = buf;
    int len = lengths[s[0] >> 3];
    unsigned char *next = s + len + !len;
    *c  = (uint32_t)(s[0] & masks[len]) << 18;
    *c |= (uint32_t)(s[1] & 0x3f) << 12;
    *c |= (uint32_t)(s[2] & 0x3f) <<  6;
    *c |= (uint32_t)(s[3] & 0x3f) <<  0;
    *c >>= shiftc[len];
    return next;
}

static void str_to_list(FL_TCB *tcb, char *str, int len, FL_VAL var, 
    FL_VAL tail)
{
    FL_VAL n = fl_nil, lst;
    char *ps = str;
    char *pe = ps + len;
    uint32_t ch;
    while(ps < pe) {
        ps = utf8_decode(ps, &ch);
        FL_VAL c = fl_alloc_cell(tcb, (FL_VAL)(((long)ch << 1) | 1), (FL_VAL)1); 
        ((FL_CELL *)c)->tag = FL_LIST_TAG;
        if(n != fl_nil) ((FL_CELL *)n)->cdr = addref(c);
        else lst = c;
        n = c;
    }
    if(n != fl_nil) ((FL_CELL *)n)->cdr = addref(tail);
    else lst = tail;
    fl_unify_result(tcb, lst, var);    
}

void fl_string_to_list_3(FL_TCB *tcb, FL_VAL str, FL_VAL var, FL_VAL tail)
{
    if(BITS(str) != 2) fl_rt_error(tcb, str, FL_NOT_AN_ATOM);
    char *ps = (char *)str;
    str_to_list(tcb, ps + 1, *ps, var, tail);
}

static unsigned char *utf8_encode(uint32_t u, unsigned char *p)
{
    if(u < 0x80) *(p++) = u;
    else if(u < 0x800) {
        *(p++) = (u >> 6) | 0xC0;
        *(p++) = (u & 0x3F) | 0x80;
    } else if(u < 0x10000) {
        *(p++) = (u >> 12) | 0xE0;
        *(p++) = ((u >> 6) & 0x3F) | 0x80;
        *(p++) = (u & 0x3F) | 0x80;
    } else if(u < 0x110000) {
        *(p++) = (u >> 18) | 0xF0;
        *(p++) = ((u >> 12) & 0x3F) | 0x80;
        *(p++) = ((u >> 6) & 0x3F) | 0x80;
        *(p++) = (u & 0x3F) | 0x80;
    }
    return p;
}

static void out_of_memory(void)
{
    fputs("out of memory\n", stderr);
    exit(1);
}

static unsigned char *cbuf = NULL;
static int cbuf_len;

static unsigned char *init_cbuf(void)
{
    if(cbuf == NULL) cbuf = malloc(cbuf_len = INIT_CBUF);
    return cbuf;
}

static unsigned char *stringify(FL_TCB *tcb, FL_VAL str, int *len)
{
    if(BITS(str) == 2) {
        if(len != NULL) *len = *((char *)str);
        return (char *)str + 1;
    }
    if(BITS(str) == 0) {
        if(TAG(str) == FL_LIST_TAG) {
            unsigned char *p = init_cbuf();
            while(BITS(str) == 0) {
                FL_VAL x = deref(CAR(str));
                if(((long)x & FL_INT_BIT) == 0) 
                    fl_rt_error(tcb, str, FL_NOT_A_STRING);
                if(p + 5 > cbuf + cbuf_len) {
                    long offset = p - cbuf;
                    cbuf = realloc(cbuf, cbuf_len *= 2);
                    if(cbuf == NULL) out_of_memory();
                    p = cbuf + offset;
                }
                p = utf8_encode(INT(x), p);
                str = deref(CDR(str));
            }
            *p = '\0';
            if(len != NULL) *len = p - cbuf;
            return cbuf;
        }
    }
    fl_rt_error(tcb, str, FL_NOT_A_STRING);
    return NULL;
}

void fl_list_to_string_2(FL_TCB *tcb, FL_VAL lst, FL_VAL var)
{
    int len;
    unsigned char *s = stringify(tcb, lst, &len);
    fl_assign(tcb, fl_mkstring(s, len), var);
}

void fl_list_to_number_3(FL_TCB *tcb, FL_VAL lst, FL_VAL base, FL_VAL var)
{
    if(((long)base & 1) != 1) fl_rt_error(tcb, base, FL_NOT_AN_INT);
    int b = INT(base);
    unsigned char *p = init_cbuf();
    int f = 0;
    FL_VAL lst0 = lst;
    while(BITS(lst) == 0) {
        FL_VAL x = deref(CAR(lst));
        if(((long)x & FL_INT_BIT) == 0) fl_rt_error(tcb, x, FL_NOT_AN_INT);
        unsigned char c = INT(x);
        if(c == '.' || c == 'e' || c == 'E') f = 1;
        *(p++) = c;
        lst = deref(CDR(lst));
    }
    *p = '\0';
    if(f) str_to_float(tcb, cbuf, var);
    else str_to_int(tcb, cbuf, b, var);
}

void fl_list_to_number_2(FL_TCB *tcb, FL_VAL lst, FL_VAL var)
{
    fl_list_to_number_3(tcb, lst, (FL_VAL)((10 << 1) | 1), var);
}

void fl_number_to_list_4(FL_TCB *tcb, FL_VAL num, FL_VAL base, FL_VAL var, 
    FL_VAL tail)
{
    char *p;
    if(((long)num & FL_INT_BIT) != 0) {
        if(((long)base & 1) != 1) fl_rt_error(tcb, base, FL_NOT_AN_INT);
        int b = INT(base);
        p = init_cbuf() + 128;
        cbuf[ 128 ] = '\0';
        long n = labs(INT(num));
        while(n != 0) {
            *(--p) = n % b + '0';
            n /= b;
        }
        if((long)num < 0) *(--p) = '-';
    } else if(BITS(num) != 0 || TAG(num) != FL_FLOAT_TAG)
        fl_rt_error(tcb, num, FL_NOT_A_NUMBER);
    else sprintf(p = init_cbuf(), "%.15g", *((double *)&((FL_CELL *)num)->car));
    str_to_list(tcb, p, strlen(p), var, tail);
}

void fl_number_to_list_3(FL_TCB *tcb, FL_VAL num, FL_VAL var, FL_VAL tail)
{
    fl_number_to_list_4(tcb, num, (FL_VAL)((10 << 1) | 1), var, tail);
}


static char fnbuf[ 256 ];

void fl_open_file_3(FL_TCB *tcb, FL_VAL name, FL_VAL mode, FL_VAL var)
{
    int len;
    char *fname = stringify(tcb, name, &len);
    strncpy(fnbuf, fname, len);
    if(BITS(mode) != 2) fl_rt_error(tcb, mode, FL_NOT_AN_ATOM);
    int m = O_RDONLY;
    mode_t mo = 420;
    switch(*((char *)mode + 1)) {
    case 'w': m = O_WRONLY | O_TRUNC | O_CREAT; break;
    case 'r': m = O_RDONLY; break;
    case 'a': m = O_WRONLY | O_APPEND; break;
    }
    int fd = open(fnbuf, m, mo);
    if(fd == -1) fl_rt_error(tcb, name, FL_IO_ERROR);
    fl_assign(tcb, (FL_VAL)(((long)fd << 1) | 1), var);
}

void fl_create_file_3(FL_TCB *tcb, FL_VAL name, FL_VAL mode, FL_VAL var)
{
    int len;
    char *fname = stringify(tcb, name, &len);
    strncpy(fnbuf, fname, len);
    if(BITS(mode) != 2) fl_rt_error(tcb, mode, FL_NOT_AN_ATOM);
    int m = O_WRONLY | O_CREAT | O_TRUNC;
    if((*((char *)mode + 1)) == 'a') m |= O_APPEND;
    int fd = open(fnbuf, m, 0644);
    if(fd == -1) fl_rt_error(tcb, name, FL_IO_ERROR);
    fl_assign(tcb, (FL_VAL)(((long)fd << 1) | 1), var);
}

void fl_close_file_2(FL_TCB *tcb, FL_VAL fd, FL_VAL var)
{
    if(((long)fd & FL_INT_BIT) == 0) fl_rt_error(tcb, fd, FL_NOT_AN_INT);
    close(INT(fd));
    fl_assign(tcb, fl_nil, var);
}

static void bytes_to_list(FL_TCB *tcb, unsigned char *str, int len, FL_VAL var, 
    FL_VAL tail)
{
    FL_VAL lst = tail;
    for(unsigned char *p = str + len - 1; p >= str; --p) {
        lst = fl_alloc_cell(tcb, (FL_VAL)(((long)*p << 1) | 1), lst);
        ((FL_CELL *)lst)->tag = FL_LIST_TAG;
    }
    fl_unify_result(tcb, lst, var);    
}

void fl_write_block_4(FL_TCB *tcb, FL_VAL fd, FL_VAL str, FL_VAL n, FL_VAL rest)
{
    if(((long)fd & FL_INT_BIT) == 0) fl_rt_error(tcb, fd, FL_NOT_AN_INT);
    int len = 0;
    if(((long)n & FL_INT_BIT) != 0) len = INT(n);
    unsigned char *p;
    if(BITS(str) == 2) {
        int slen = *((char *)str);
        p = (char *)str + 1;
        if(len == 0 || len > slen) len = slen;
        int w = write(INT(fd), p, len);
        if(w == -1) {
            if(errno == EINTR) fl_assign(tcb, fl_false, rest);
            else fl_rt_error(tcb, fd, FL_IO_ERROR);
            return;
        }
        if(w < slen) bytes_to_list(tcb, p + w, slen - w, rest, fl_nil);
        else fl_assign(tcb, fl_nil, rest);
        return;
    }
    if(TAG(str) != FL_LIST_TAG) fl_rt_error(tcb, str, FL_NOT_A_STRING);
    p = init_cbuf();
    FL_VAL lst = str;
    int count = 0;
    while(BITS(lst) == 0) {
        if(p >= cbuf + cbuf_len) {
            int offset = p - cbuf;
            cbuf = realloc(cbuf, cbuf_len *= 2);
            if(cbuf == NULL) out_of_memory();
            p = cbuf + offset;
        }
        FL_VAL x = CAR(lst);
        if(((long)x & FL_INT_BIT) == 0) fl_rt_error(tcb, lst, FL_NOT_A_STRING);
        *(p++) = INT(x);
        lst = deref(CDR(lst));
        ++count;
        if(len && count >= len) break;
    }
    int w = write(INT(fd), cbuf, len ? len : count);
    if(w < len) {
        while(w-- && BITS(str) == 0) str = deref(CDR(str));
        fl_unify_result(tcb, str, rest);
    } else fl_unify_result(tcb, lst, rest);
}

void fl_read_block_4(FL_TCB *tcb, FL_VAL fd, FL_VAL n, FL_VAL lst, FL_VAL tail)
{   
    if(((long)fd & FL_INT_BIT) == 0) fl_rt_error(tcb, fd, FL_NOT_AN_INT);
    if(((long)n & FL_INT_BIT) == 0) fl_rt_error(tcb, n, FL_NOT_AN_INT);
    int len = INT(n);
    init_cbuf();
    if(len >= cbuf_len) len = cbuf_len;
    int r = read(INT(fd), cbuf, len);
    if(r == -1) {
        if(errno == EINTR) fl_assign(tcb, fl_false, lst);
        else fl_rt_error(tcb, fd, FL_IO_ERROR);
    } else if(r == 0) fl_assign(tcb, fl_nil, lst);
    else bytes_to_list(tcb, cbuf, r, lst, tail);
}

void fl_utf_encode_3(FL_TCB *tcb, FL_VAL ch, FL_VAL lst, FL_VAL tail)
{
    if(((long)ch & FL_INT_BIT) == 0) fl_rt_error(tcb, ch, FL_NOT_AN_INT);
    unsigned char *p = utf8_encode(INT(ch), init_cbuf());
    bytes_to_list(tcb, cbuf, p - cbuf, lst, tail);
}

void fl_utf_decode_3(FL_TCB *tcb, FL_VAL lst, FL_VAL var, FL_VAL rest)
{
    int n = 1;
    FL_VAL ch = CAR(lst);
    if(((long)ch & FL_INT_BIT) == 0) fl_rt_error(tcb, ch, FL_NOT_AN_INT);
    unsigned char c = INT(ch);
    if((c & 0xf8) == 0xf0) n = 4;
    else if((c & 0xf0) == 0xe0) n = 3;
    else if((c & 0xe0) == 0xc0) n = 2;
    unsigned char *p = init_cbuf();
    *(p++) = c;
    lst = deref(CDR(lst));
    while(--n > 0) {
        if(BITS(lst) != 0) {
            while(n--) *(p++) = 0;
            break;
        }
        ch = CAR(lst);
        if(((long)ch & FL_INT_BIT) == 0) fl_rt_error(tcb, ch, FL_NOT_AN_INT);
        *(p++) = INT(ch);
        lst = deref(CDR(lst));
    }
    uint32_t uc;
    utf8_decode(init_cbuf(), &uc);
    fl_assign(tcb, (FL_VAL)(((long)uc << 1) | 1), var);
    fl_unify_result(tcb, lst, rest);
}

void fl_send_2(FL_TCB *tcb, FL_VAL port, FL_VAL x)
{
    fl_send_to_port(tcb, port, x);
}

void fl_send_3(FL_TCB *tcb, FL_VAL port, FL_VAL x, FL_VAL done)
{
    fl_send_2(tcb, port, x);
    fl_assign(tcb, fl_nil, done);
}

void fl_open_port_2(FL_TCB *tcb, FL_VAL pv, FL_VAL sv)
{
    FL_VAL port = fl_alloc_cell(tcb, sv, fl_nil);
    ((FL_CELL *)port)->tag = FL_PORT_TAG;
    fl_assign(tcb, port, pv);
}

void fl_list_to_tuple_2(FL_TCB *tcb, FL_VAL lst, FL_VAL var)
{
    if(lst == fl_nil) fl_assign(tcb, fl_nil, var);
    else {
        int len = 0;
        FL_VAL n = deref(CDR(lst)), args, *prev = NULL;
        while(BITS(n) == 0) {
            FL_VAL arg = mklist(tcb, deref(CAR(n)), fl_nil);
            if(prev == NULL) args = arg;
            else *prev = addref(arg);
            prev = &CDR(arg);
            n = deref(CDR(n));
            ++len;
        }
        FL_VAL t = mktuple(tcb, deref(CAR(lst)), len, args);
        fl_unify_result(tcb, t, var);
    }
}

void fl_tuple_to_list_3(FL_TCB *tcb, FL_VAL tup, FL_VAL var, FL_VAL tail)
{
    if(tup == fl_nil) fl_assign(tcb, var, tail);
    else if(BITS(tup) != 0 || (((FL_CELL *)tup)->tag & FL_TUPLE_BIT) == 0)
        fl_rt_error(tcb, tup, FL_NOT_A_TUPLE);
    else {
        FL_VAL n = fl_alloc_cell(tcb, CAR(tup), fl_nil);
        FL_VAL lst = n;
        ((FL_CELL *)lst)->tag = FL_LIST_TAG;
        tup = CDR(tup);
        if(tail == fl_nil) ((FL_CELL *)lst)->cdr = addref(tup);
        else {
            while(BITS(tup) == 0) {
                FL_VAL n2 = fl_alloc_cell(tcb, CAR(tup), fl_nil);
                ((FL_CELL *)n2)->tag = FL_LIST_TAG;
                ((FL_CELL *)n)->cdr = addref(n2);
                tup = deref(CDR(tup));
                n = n2;
            }
            ((FL_CELL *)n)->cdr = addref(tail);
        }
        fl_unify_result(tcb, lst, var);
    }
}

void fl_put_arg_3(FL_TCB *tcb, FL_VAL idx, FL_VAL tup, FL_VAL x)
{
    if(BITS(tup) != 0 || (((FL_CELL *)tup)->tag & FL_TUPLE_BIT) == 0)
        fl_rt_error(tcb, tup, FL_NOT_A_TUPLE);
    int len = (((FL_CELL *)tup)->tag & ~FL_TUPLE_BIT) >> 24;
    if(((long)idx & FL_INT_BIT) == 0) fl_rt_error(tcb, idx, FL_NOT_AN_INT);
    int i = INT(idx) - 1;
    if(i < 0 || i > len) fl_rt_error(tcb, idx, FL_BAD_INDEX);
    while(i--) tup = deref(CDR(tup));
    fl_unify_result(tcb, x, CAR(tup));
}

void fl_put_arg_4(FL_TCB *tcb, FL_VAL idx, FL_VAL tup, FL_VAL x, FL_VAL done)
{
    fl_put_arg_3(tcb, idx, tup, x);
    fl_assign(tcb, fl_nil, done);
}

void fl_get_arg_3(FL_TCB *tcb, FL_VAL idx, FL_VAL tup, FL_VAL var)
{
    if(BITS(tup) != 0 || (((FL_CELL *)tup)->tag & FL_TUPLE_BIT) == 0)
        fl_rt_error(tcb, tup, FL_NOT_A_TUPLE);
    int len = (((FL_CELL *)tup)->tag & ~FL_TUPLE_BIT) >> 24;
    if(((long)idx & FL_INT_BIT) == 0) fl_rt_error(tcb, idx, FL_NOT_AN_INT);
    int i = INT(idx) - 1;
    if(i < 0 || i > len) fl_rt_error(tcb, idx, FL_BAD_INDEX);
    while(i--) tup = deref(CDR(tup));
    fl_unify_result(tcb, CAR(tup), var);
}

void fl_get_arg_4(FL_TCB *tcb, FL_VAL idx, FL_VAL tup, FL_VAL var, FL_VAL done)
{
    fl_get_arg_3(tcb, idx, tup, var);
    fl_assign(tcb, fl_nil, done);
}

void fl_make_tuple_2(FL_TCB *tcb, FL_VAL n, FL_VAL var)
{
    if(((long)n & FL_INT_BIT) == 0) fl_rt_error(tcb, n, FL_NOT_AN_INT);
    int len = INT(n);
    if(len < 0 || len > 0x7f) fl_rt_error(tcb, n, FL_BAD_INDEX);
    if(len == 0) fl_assign(tcb, fl_nil, var);
    else {
        FL_VAL tup = fl_nil, prev;
        int i = len;
        while(i--) {
            FL_VAL v = fl_alloc_cell(tcb, fl_nil, fl_nil);
            ((FL_CELL *)v)->tag = FL_VAR_TAG;
            ((FL_CELL *)v)->car = v;
            FL_VAL n = fl_alloc_cell(tcb, v, fl_nil);
            if(tup == fl_nil) {
                tup = n;
                ((FL_CELL *)tup)->tag = FL_TUPLE_BIT | ((len - 1) << 24);
            } else {
                ((FL_CELL *)n)->tag = FL_LIST_TAG;
                ((FL_CELL *)prev)->cdr = addref(n);
            }
            prev = n;
        }
        fl_unify_result(tcb, tup, var);
    }
}

void fl_halt_1(FL_TCB *tcb, FL_VAL status)
{
    if(((long)status & FL_INT_BIT) == 0) 
        fl_rt_error(tcb, status, FL_NOT_AN_INT);
    exit(INT(status));
}

void fl_events_1(FL_TCB *tcb, FL_VAL stream)
{
    tcb->events = addref(stream);
}

void fl_forward_3(FL_TCB *tcb, FL_VAL id, FL_VAL msg, FL_VAL done)
{
    int peer = INT(id);
    if(peer < 0 && peer >= FL_WEST) peer = tcb->peers[ -peer - 1 ];
    else if(peer < 1 || peer > fl_nthreads()) fl_rt_error(tcb, id, FL_BAD_PEER);
    if(fl_forward(tcb, peer, msg)) fl_assign(tcb, MKINT(1), done);
    else fl_assign(tcb, MKINT(0), done);
}

void fl_local_value_addref_1(FL_TCB *tcb, FL_VAL x)
{
    /* addref represented value of RPTR: */    
    x = RPTR(x);
    addref(x);
}

void fl_unref_1(FL_TCB *tcb, FL_VAL x)
{
    unref(tcb, x);
}

void fl_local_value_2(FL_TCB *tcb, FL_VAL rptr, FL_VAL val)
{
    /* convert RPTR into represented value: */
    fl_assign(tcb, RPTR(rptr), val);
}

void fl_local_unref_1(FL_TCB *tcb, FL_VAL rptr)
{
    /* unref represented value of RPTR: */
    FL_VAL x = RPTR(rptr);
    unref(tcb, x);
}

void fl_local_value_unref_2(FL_TCB *tcb, FL_VAL rptr, FL_VAL val)
{
    /* convert RPTR into represented value and unref it: */
    fl_assign(tcb, RPTR(rptr), val);
    unref(tcb, RPTR(rptr));
}

void fl_remote_id_2(FL_TCB *tcb, FL_VAL rptr, FL_VAL id)
{
    fl_assign(tcb, MKINT(RPTR_ID(rptr)), id);
}

void fl_local_assign_2(FL_TCB *tcb, FL_VAL val, FL_VAL rptr)
{
    FL_VAL var = RPTR(rptr);
    fl_assign(tcb, val, var);
    unref(tcb, var);
}

void fl_call_event_3(FL_TCB *tcb, FL_VAL peer, FL_VAL term, FL_VAL mod)
{
    FL_VAL ev1 = fl_alloc_cell(tcb, term, mod);
    ((FL_CELL *)ev1)->tag = FL_LIST_TAG;
    FL_VAL ev2 = fl_alloc_cell(tcb, peer, ev1);
    ((FL_CELL *)ev2)->tag = FL_LIST_TAG;
    FL_VAL ev = fl_alloc_cell(tcb, MKINT(6), ev2); /* SEND_CALL */
    ((FL_CELL *)ev)->tag = FL_LIST_TAG;
    fl_add_event(tcb, ev);
}

void fl_assign_3(FL_TCB *tcb, FL_VAL dest, FL_VAL src, FL_VAL var)
{
    fl_assign(tcb, src, dest);
    fl_assign(tcb, fl_nil, var);
}

void fl_listen_2(FL_TCB *tcb, FL_VAL fd, FL_VAL var)
{
    if((BITS(fd) & 1) == 0) fl_rt_error(tcb, fd, FL_NOT_AN_INT);
    if(fl_add_event_to_queue(tcb, FL_INPUT, fd, var) == -1)
        fl_rt_error(tcb, fd, FL_IO_ERROR);
}

void fl_mkmodule_2(FL_TCB *tcb, FL_VAL mod, FL_VAL var)
{
    FL_VAL m = fl_alloc_cell(tcb, mod, fl_nil);
    ((FL_CELL *)m)->tag = FL_MODULE_TAG;
    fl_assign(tcb, m, var);
}

void fl_threads_1(FL_TCB *tcb, FL_VAL var)
{
    fl_assign(tcb, (FL_VAL)(((long)fl_nthreads() << 1) | 1), var);
}

void fl_execute_4(FL_TCB *tcb, FL_VAL args, FL_VAL fds, FL_VAL rpid, FL_VAL status)
{
    /* fds = [In, Out, Err, Closelist] */
    int fdi = INT(deref(CAR(fds)));
    fds = deref(CDR(fds));
    int fdo = INT(deref(CAR(fds)));
    fds = deref(CDR(fds));
    int fde = INT(deref(CAR(fds)));
    fds = deref(CDR(fds));
    int cfds[ 16 ], n = 0;
    for(fds = deref(CAR(fds)); BITS(fds) == 0; fds = deref(CDR(fds)))
        cfds[ n++ ] = INT(deref(CAR(fds)));
    char *argv[ MAX_ARGS ];
    int c = 0;
    FL_VAL ap = args;
    init_cbuf();
    while(BITS(args) == 0) {
        char *arg = stringify(tcb, deref(CAR(args)), NULL);
        argv[ c++ ] = strdup(arg);
        args = deref(CDR(args));
    }
    argv[ c ] = NULL;
    int pid = fork();
    if(pid == -1) fl_rt_error(tcb, fl_nil, FL_IO_ERROR);
    if(pid == 0) {
        if(fdi != STDIN_FILENO) {
            dup2(fdi, STDIN_FILENO);
            close(fdi);
        }
        if(fdo != STDOUT_FILENO) {
            dup2(fdo, STDOUT_FILENO);
            close(fdo);
        }
        if(fde != STDERR_FILENO) {
            dup2(fde, STDERR_FILENO);
            close(fde);
        }
        for(int i = 0; i < n; ++i) close(cfds[ i ]);
        execvp(argv[ 0 ], argv);
        fprintf(stderr, "execve failed: %s\n", strerror(errno));
        exit(1);
    } 
    /* parent */
    FL_VAL vpid = (FL_VAL)(((long)pid << 1) | 1);
    fl_assign(tcb, vpid, rpid);
    if(fl_add_event_to_queue(tcb, FL_CHILD, vpid, status) == -1)
        fl_rt_error(tcb, vpid, FL_IO_ERROR);
}

void fl_open_pipe_2(FL_TCB *tcb, FL_VAL in, FL_VAL out)
{
    int fds[ 2 ];
    if(pipe(fds) == -1) fl_rt_error(tcb, fl_nil, FL_IO_ERROR);
    fl_assign(tcb, (FL_VAL)(((long)fds[ 0 ] << 1) | 1), in);
    fl_assign(tcb, (FL_VAL)(((long)fds[ 1 ] << 1) | 1), out);
}

void fl_getpid_1(FL_TCB *tcb, FL_VAL pid)
{
    fl_assign(tcb, MKINT(getpid()), pid);
}

void fl_self_1(FL_TCB *tcb, FL_VAL id)
{
    fl_assign(tcb, MKINT(tcb->ordinal), id);
}

void fl_current_seconds_1(FL_TCB *tcb, FL_VAL var)
{
    fl_assign(tcb, MKINT(time(NULL)), var);
}

void fl_timer_2(FL_TCB *tcb, FL_VAL ms, FL_VAL var)
{
    fl_add_event_to_queue(tcb, FL_TIMEOUT, ms, var);
}

void fl_signal_3(FL_TCB *tcb, FL_VAL sig, FL_VAL var, FL_VAL done)
{
    fl_add_event_to_queue(tcb, FL_SIGNAL, sig, var);   
    fl_assign(tcb, fl_nil, done);
}

void fl_random_1(FL_TCB *tcb, FL_VAL var)
{
    fl_assign(tcb, fl_alloc_float(tcb, fl_random()), var);
}

void fl_set_random_seed_2(FL_TCB *tcb, FL_VAL seed, FL_VAL done)
{
    int len;
    unsigned char *ptr = stringify(tcb, seed, &len);
    fl_set_random_seed(ptr, len);
    fl_assign(tcb, fl_nil, done);
}
