% FLENG compiler - x86_64 backend
%
% register usage:
%
%   A: rax
%   T: rbx (callee save)
%   C: rdi
%   G: r12 (callee save)
%   N: r13 (callee save)
%   E0-E1: r14-r15 (callee save)
%   E<n>: (n >= 2) stack, F points to E2, ...
%   FP: rbp
%   P0-P3: rsi, rdx, rcx, r8 (tcb is passed in rdi)
%   SP: rsp, on entry set to rbp
%   X0, X1: r9, r10

:- dynamic([external/1]).

d_setup(_).
d_finalize(_).

emit_external(Name, _) :- external(Name), !.
emit_external(Name, _) :- entry_point(_, Name), !.
emit_external(Name, Out) :-
    assertz(external(Name)),
    d_extern(Name, Out).

d_extern(Name, Out) :- emit('  extern ~a\n', [Name], Out).
d_global(Name, Out) :- emit('  global ~a\n', [Name], Out).
d_section(Name, Out) :- emit('  SECTION ~a\n', [Name], Out).
d_comment(Str, Out) :- emit('; ~a\n', [Str], Out).
d_wreserve(N, Out) :- emit('  resq ~d\n', [N], Out).

d_function_entry(Out) :- emit('  push rbp\n', Out).
d_function_exit(Out) :- emit('  pop rbp\n', Out).
d_endfunction(_).

c_prepare_foreign_call(_, _).
c_finalize_foreign_call(_, _).

d_align(code, Out) :- emit('  align 16\n', Out).
d_align(even, Out) :- emit('  align 2\n', Out).
d_align(halfword, Out) :- emit('  align 4\n', Out).
d_align(word, Out) :- emit('  align 8\n', Out).
d_align(float, Out) :- emit('  align 8\n', Out).

d_equ(S, X, Out) :- emit('~a equ ~d\n', [S, X], Out).
d_equ(S, Index, X, Out) :- emit('~a~d equ ~d\n', [S, Index, X], Out).

d_bdata([B], Out) :- !, emit('  db ~w\n', [B], Out).
d_bdata([B|L], Out) :- 
    emit('  db ~w', [B], Out),
    d_data1(L, Out).

d_wdata([B], Out) :- !, emit('  dq ~w\n', [B], Out).
d_wdata([B|L], Out) :- 
    emit('  dq ~w', [B], Out),
    d_data1(L, Out).

d_fdata(D, Out) :- d_wdata(D, Out).

d_data1([], Out) :- emit('\n', Out).
d_data1([B|L], Out) :- 
    emit(', ~w', [B], Out),
    d_data1(L, Out).

d_label(Str, Out) :- emit('~a:\n', [Str], Out).
d_label(Prefix, Index, Out) :- emit('~a~d:\n', [Prefix, Index], Out).

a_call(Name, Out) :- emit('  call ~a\n', [Name], Out).

r_reg(a, rax).
r_reg(c, rdi).
r_reg(t, rbx).
r_reg(g, r12).
r_reg(n, r13).
r_reg(fp, rbp).
r_reg(sp, rsp).
r_reg(e(N), R) :- e_reg(N, R).
r_reg(p(N), R) :- p_reg(N, R).
r_reg(x(N), R) :- x_reg(N, R).

p_reg(0, rsi).
p_reg(1, rdx).
p_reg(2, rcx).
p_reg(3, r8).

x_reg(0, r9).
x_reg(1, r10).

e_reg(0, r14).
e_reg(1, r15).

r_move(Src, e(N), Out) :- 
    N > 1, 
    !,
    N2 is N - 2,
    r_store(Src, fp, N2, Out).
r_move(e(N), Dest, Out) :- 
    N > 1, 
    !,
    N2 is N - 2,
    r_load(fp, N2, Dest, Out).
r_move(Src, Dest, Out) :-
    r_reg(Src, SR),
    r_reg(Dest, DR),
    emit('  mov ~a, ~a\n', [DR, SR], Out).

r_load(RI, I, RD, Out) :-
    r_reg(RI, RIR),
    r_reg(RD, RDR),
    emit('  mov ~a, [~a + ~d * 8]\n', [RDR, RIR, I], Out).

r_compare(RS, RI, I, Out) :-
    r_reg(RI, RIR),
    r_reg(RS, RSR),
    emit('  cmp [~a + ~d * 8], ~a\n', [RIR, I, RSR], Out).

a_load(A, I, RD, Out) :-
    r_reg(RD, RDR),
    emit('  mov ~a, [~a + ~d * 8]\n', [RDR, A, I], Out).

r_store(RS, RI, I, Out) :-
    r_reg(RI, RIR),
    r_reg(RS, RSR),
    emit('  mov [~a + ~d * 8], ~a\n', [RIR, I, RSR], Out).

a_store(RS, A, I, Out) :-
    r_reg(RS, RSR),
    emit('  mov [~a + ~d * 8], ~a\n', [A, I, RSR], Out).

la_load(L, RD, Out) :-
    r_reg(RD, RDR),
    emit('  mov ~a, ~a\n', [RDR, L], Out).
la_load(L, Index, RD, Out) :-
    r_reg(RD, RDR),
    emit('  mov ~a, ~a~d\n', [RDR, L, Index], Out).

i_load(I, RD, Out) :-
    r_reg(RD, RDR),
    emit('  mov ~a, ~w\n', [RDR, I], Out).

f_load(F, Out) :-
    emit('  mov rdx, __float64__(~f)\n', [F], Out).

r_push(R, Out) :-
    r_reg(R, RR),
    emit('  push ~a\n', [RR], Out).

r_pop(R, Out) :-
    r_reg(R, RR),
    emit('  pop ~a\n', [RR], Out).

g_return(Out) :- emit('  ret\n', Out).

ra_inc(R, Out) :-
    r_reg(R, RR),
    emit('  inc dword [~a]\n', [RR], Out).

la_inc(L, Out) :- emit('  inc dword [~a]\n', [L], Out).

sp_align(Out) :- emit('  and rsp, -16\n', Out).

ra_add(RS, N, RD, Out) :-
    r_reg(RS, RSR),
    r_reg(RD, RDR),
    emit('  lea ~a, [~a + ~d * 8]\n', [RDR, RSR, N], Out).

r_r_op(OP, RA, RB, RD, Out) :- 
    r_reg(RA, RAR),
    r_reg(RB, RBR),
    emit('  ~a ~a, ~a\n', [OP, RAR, RBR], Out),
    r_move(RA, RD, Out).

r_and(RA, RB, RD, Out) :- r_r_op(and, RA, RB, RD, Out).
r_or(RA, RB, RD, Out) :- r_r_op(or, RA, RB, RD, Out).
r_xor(RA, RB, RD, Out) :- 
    % special case
    r_reg(RA, RAR),
    r_reg(RB, RBR),
    r_reg(RD, RDR),
    emit('  xor ~a, ~a\n  lea ~a, [~a + 1]\n', [RAR, RBR, RDR, RAR], Out).

i_test(R, I, Out) :-
    r_reg(R, RR),
    emit('  test ~a, ~d\n', [RR, I], Out).

i_mask(R, M, Out) :-
    r_reg(R, RR),
    emit('  and ~a, ~w\n', [RR, M], Out).

i_or(M, R, Out) :-
    r_reg(R, RR),
    emit('  or ~a, ~w\n', [RR, M], Out).

j_if_carry(L, Out) :- emit('  jc ~a\n', [L], Out).
j_if_carry(L, Index, Out) :- emit('  jc ~a~d\n', [L, Index], Out).

j_if_equal(L, Out) :- emit('  je ~a\n', [L], Out).
j_if_equal(L, Index, Out) :- emit('  je ~a~d\n', [L, Index], Out).

j_if_not_equal(L, Out) :- emit('  jne ~a\n', [L], Out).
j_if_not_equal(L, Index, Out) :- emit('  jne ~a~d\n', [L, Index], Out).

j_always(L, Out) :- emit('  jmp ~a\n', [L], Out).
j_always(P, Index, Out) :- emit('  jmp ~a~d\n', [P, Index], Out).
